//
// File:    hm_MakeMake.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_MakeMake
#include "hm_MakeMake.h"
#endif

#ifndef HDR_hm_Host
#include "hm_Host.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

#if HOST==RISC_OS
#ifndef LIB_kernel
#include "kernel.h"
#define LIB_kernel
#endif
#endif

// C++

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_list
#include <list>
#define LIB_list
#endif

#ifndef LIB_fstream
#include <fstream>
#define LIB_fstream
#endif

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef HDR_le_BMatrix
#include "le_BMatrix.h"
#endif

// HMake

#ifndef HDR_hm_Component
#include "hm_Component.h"
#endif

#ifndef HDR_hm_Components
#include "hm_Components.h"
#endif

#ifndef HDR_hm_Args
#include "hm_Args.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_makemake
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

// -1: nicht vorhanden

int hm_makemake::GetIndex (const string& Name) const
{
    int Index = 0;
    for (vector<hm_component>::const_iterator Iter (d_Components.begin());
        Iter!=d_Components.end(); ++Iter, ++Index)
        if ((*Iter).GetName()==Name)
            return Index;

    return -1;
}

void hm_makemake::CreateFile (const string& MakePath)
{
    delete d_Makefile;
    d_Makefile = new ofstream (MakePath.c_str());
}

void hm_makemake::WriteMacros()
{
    assert (d_Makefile);

    *d_Makefile << "# Tools and toolflags:" << endl
                << endl
                << "CppFlags = -c" << endl
                << "CppTool = g++" << endl
                << "LinkFlags = -o $@" << endl
                << "LinkTool = g++" << endl
                << "LIBS = ";

    *d_Makefile << endl << endl;
}

void hm_makemake::BuildIncludeMatrix()
{
    delete d_Include;

    d_Include = new le_binary_matrix (d_Components.size());
    for (int i=d_Components.size()-1; i>=0; --i) d_Include->Add();
    d_Include->Clear();

    int Index = 0;
    for (vector<hm_component>::const_iterator Iter (d_Components.begin());
        Iter!=d_Components.end(); ++Iter, ++Index)
    {
        for (list<string>::const_iterator Iter2 ((*Iter).GetIncludes (false).begin());
            Iter2!=(*Iter).GetIncludes (false).end(); ++Iter2)
        {
            int Index2 = GetIndex (hm_component::StripInclude (*Iter2));
            if (Index2!=-1) d_Include->Set (Index, Index2);
        }

        for (list<string>::const_iterator Iter2 ((*Iter).GetIncludes (true).begin());
            Iter2!=(*Iter).GetIncludes (true).end(); ++Iter2)
        {
            int Index2 = GetIndex (hm_component::StripInclude (*Iter2));
            if (Index2!=-1) d_Include->Set (Index, Index2);
        }
    }

    d_Include->MakeTransitive();
}

void hm_makemake::WriteFinalTarget()
{
    assert (d_Makefile);

    *d_Makefile << "# Final targets:" << endl
                << endl
                << "all : ";

    for (vector<hm_component>::const_iterator Iter (d_Components.begin());
        Iter!=d_Components.end(); ++Iter)
        if ((*Iter).HasMain())
            *d_Makefile << " \\" << endl << '\t' << (*Iter).GetName();

    *d_Makefile << endl << endl;
}

void hm_makemake::WriteTarget (int Index)
{
    assert (d_Makefile);
    assert (d_Include);

    const hm_component& Target = d_Components[Index];

    d_Clean += " ";
    d_Clean += Target.GetName();

    *d_Makefile << Target.GetName();
#if HOST==RISC_OS
    *d_Makefile << "Lib = o." << Target.GetName();
#else
    *d_Makefile << "Lib = " << Target.GetName() << ".o";
#endif

    for (int i=0; i<d_Include->GetSize(); ++i)
        if (i!=Index && d_Include->Get (Index, i))
#if HOST==RISC_OS
            *d_Makefile << " o." << d_Components[i].GetName();
#else
            *d_Makefile << " " << d_Components[i].GetName() << ".o";
#endif

    *d_Makefile << endl
                << Target.GetName() << ": $(LIBS) $(" << Target.GetName() << "Lib) " << endl
                << "\t$(LinkTool) $(LinkFlags) $(LIBS) $(" << Target.GetName()
                << "Lib) " << endl
                << endl;
}

void hm_makemake::WriteSuffixes()
{
    assert (d_Makefile);

    *d_Makefile << "# Implicit rules" << endl
                << endl
                << ".SUFFIXES: ." << d_Args.GetCpp() << endl
                << endl
                << '.' << d_Args.GetCpp() << ".o:; $(CppTool) $(CppFlags) $<"
                << endl
                << endl;
}

void hm_makemake::Finish (const string& MakePath)
{
    delete d_Makefile;
    d_Makefile = 0;

#if HOST==RISC_OS
    string Command ("SetType ");
    Command += MakePath;
    Command += " fe1";
    _kernel_oscli (Command.c_str());
#endif

    delete d_Include;
    d_Include = 0;

    d_Components.clear();
}

// Konstruktoren ---------------------------------------------------------------------------

hm_makemake::hm_makemake (const hm_args& Args)
: d_Args (Args), d_Makefile (0), d_Include (0)
{}

hm_makemake::~hm_makemake()
{
    delete d_Makefile;
    delete d_Include;
}

// Nicht-Konstante -------------------------------------------------------------------------

// Return: Ok?

bool hm_makemake::Make (const hm_components& Components)
{
    d_Clean.clear();
    
    // Umkopieren
    d_Components.clear();
    d_Components.reserve (Components.GetComponents().size());

    for (list<hm_component>::const_iterator Iter (Components.GetComponents().begin());
        Iter!=Components.GetComponents().end(); ++Iter)
    {
        d_Components.push_back (*Iter);
#if HOST==RISC_OS
    	d_Clean += " o.";
	d_Clean += Iter->GetName();
#else
    	d_Clean += " ";
	d_Clean += Iter->GetName();
	d_Clean += ".o";
#endif	
    }

    // Makefile
    string MakePath ("Makefile");
    CreateFile (MakePath);

    // Makros
    WriteMacros();

    // Matrix aufstellen
    BuildIncludeMatrix();

    // Final Target
    int N = 0;
    for (vector<hm_component>::const_iterator Iter (d_Components.begin());
        Iter!=d_Components.end(); ++Iter)
        if ((*Iter).HasMain())
            ++N;

    if (N==0) cerr << "Warning: no main" << endl;

    if (N>1) WriteFinalTarget();

    // Targets
    int Index = 0;
    for (vector<hm_component>::const_iterator Iter (d_Components.begin());
        Iter!=d_Components.end(); ++Iter, ++Index)
        if ((*Iter).HasMain())
            WriteTarget (Index);

    // Clean
    *d_Makefile << "clean:" << endl << "\trm" << d_Clean << endl << endl;    

    // Suffixes
    WriteSuffixes();

    bool Ok = d_Makefile->good();
    Finish (MakePath);
    return Ok;
}
