//
// File:    hm_SpecScan.c++
// Author:  Marc Zinnschlag

#ifndef HDR_hm_SpecScan
#include "hm_SpecScan.h"
#endif

// C

#ifndef LIB_cassert
#include <cassert>
#define LIB_cassert
#endif

#ifndef LIB_cctype
#include <cctype>
#define LIB_cctype
#endif

// C++

#ifndef LIB_algorithm
#include <algorithm>
#define LIB_algorithm
#endif

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

// -----------------------------------------------------------------------------------------
// Klasse: hm_spec_line
// -----------------------------------------------------------------------------------------

// Konstruktoren ---------------------------------------------------------------------------

hm_spec_line::hm_spec_line() : d_Type (NoFile) {}

// File: Anteil links des Doppelpunkt ohne '.' und Endung

hm_spec_line::hm_spec_line (const string& File, file_type Type)
: d_File (File), d_Type (Type)
{
    assert (!File.empty());
}

hm_spec_line::~hm_spec_line() {}

// Nicht-Konstante -------------------------------------------------------------------------

// Argument hinzufgen (Anteil rechts des Doppelpunkt)

void hm_spec_line::AddArg (const string& Arg)
{
    assert (!Arg.empty());
    d_Args.push_back (Arg);
}

// Freie Operatoren ------------------------------------------------------------------------

ostream& operator<< (ostream& Stream, const hm_spec_line& Line)
{
    // Linker Anteil
    Stream << Line.GetFile();

    switch (Line.GetType())
    {
        case hm_spec_line::CFile: Stream << ".c"; break;
        case hm_spec_line::HFile: Stream << ".h"; break;
     // case hm_spec_line::NoFile: Klasse, daher keine Endung
    }

    Stream << ": ";

    // Rechter Anteil
    bool First = true;
    for (list<string>::const_iterator Iter (Line.GetArgs().begin());
        Iter!=Line.GetArgs().end(); ++Iter)
    {
        if (First) First = false; else Stream << ' ';

        Stream << *Iter;
    }

    return Stream;
}

// -----------------------------------------------------------------------------------------
// Klasse: hm_spec_scanner
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

void hm_spec_scanner::ReportError (const string& Error)
{
    cerr << "Line " << d_Line << ": " << Error << endl;
    d_Errors = true;
}

// Zum Anfang der nchsten logischen Zeile vorrcken

void hm_spec_scanner::SkipLine()
{
    char c;
    while (GetNext (c) && c!='\n');
}

// Ungefiltertes Zeichen lesen und ggf. Zeilennummer erhhen

bool hm_spec_scanner::ReadChar (char& c)
{
    bool Result = d_Stream.get (c);

    if (Result)
    {
        if (d_NewLine)
        {
            ++d_Line;
            d_NewLine = false;
        }

        if (c=='\n') d_NewLine = true;
    }

    return Result;
}

void hm_spec_scanner::Putback (char c)
{
    if (c=='\n') d_NewLine = false;
    d_Stream.putback (c);
}

// Return: Konnte Zeichen gelesen werden (sonst c undef)
// \-Zeilenumbrche und Kommentare bereits herausgefiltert

bool hm_spec_scanner::GetNext (char& c)
{
    for (;;)
    {
        if (!ReadChar (c)) return false;

        if (c=='#')
        {
            // Kommentar: Bis zum Ende der Zeile weiterlaufen
            c = 0;
            while (ReadChar (c) && c!='\n');
            return d_Stream.good();
        }
        else if (c=='\\')
        {
            // Verbindung zweier physikalischen Zeilen
            bool Error = false;
            while (ReadChar (c) && c!='\n')
            {
                if (!IsWhitespace (c) && !Error)
                {
                    ReportError ("garbage at end of line (past '\\')");
                    Error = true;
                }
            }
        }
        else
            return true;
    }
}

void hm_spec_scanner::SkipWhitespaces (bool SkipNewLine)
{
    char c;

    while (ReadChar (c) && (IsWhitespace (c) || SkipNewLine && c=='\n'));
    if (d_Stream.good()) Putback (c);
}

// Return: Ok?

bool hm_spec_scanner::ReadFile (string& File)
{
    File = "";

    char c;
    while (GetNext (c))
    {
        if (c!='.' && c!=':' && !IsWhitespace (c))
            File += c;
        else
            break;
    }

    if (!d_Stream.good()) return false;

    if (IsWhitespace (c))
    {
        SkipWhitespaces();
        if (!GetNext (c))
            return false;
    }

    if (c=='\n') return false;

    Putback (c);

    return true;
}

// Return: Ok?

bool hm_spec_scanner::ReadType (hm_spec_line::file_type& Type)
{
    char c;
    if (!GetNext (c)) return false;

    if (c==':')
    {
        Type = hm_spec_line::NoFile;
        Putback (c);
        return true;
    }

    if (c!='.')
    {
        SkipLine();
        return false;
    }

    string TypeString;

    while (GetNext (c) && c!='\n' && c!=':')
        TypeString += c;

    if (c=='\n') return false;
    Putback (c);

    for (int i=TypeString.length()-1; i>=0; --i)
        TypeString[i] = tolower (TypeString[i]);

    if (TypeString=="h")
        Type = hm_spec_line::HFile;
    else if (TypeString=="c" || TypeString=="cc" || TypeString=="cpp" || TypeString=="c++")
        Type = hm_spec_line::CFile;
    else
    {
        ReportError ("." + TypeString + " \?\?\?");
        return false;
    }

    SkipWhitespaces();
    return true;
}

// Falls kein Argument mehr gelesen werden konnte, wird ein leerer String zurckgegeben

string hm_spec_scanner::ReadArg()
{
    SkipWhitespaces();

    string Arg;

    char c;
    while (GetNext (c))
    {
        if ( c=='\n' || IsWhitespace (c))
        {
            Putback (c);
            break;
        }

        Arg += c;
    }

    return Arg;
}

// Konstruktoren ---------------------------------------------------------------------------

// Stream: Spec-Datei (Stream mu mindestens so lange ex. wie *this in Gebrauch ist)

hm_spec_scanner::hm_spec_scanner (istream& Stream, const list<string>& LegalArgs)
: d_Stream (Stream), d_Line (1), d_NewLine (false), d_Errors (false),
  d_LegalArgs (LegalArgs)
{}

hm_spec_scanner::~hm_spec_scanner() {}

// Nicht-Konstante -------------------------------------------------------------------------

// Logische Zeile einlesen
// Return: Konnte Zeile eingelesen werden (sonst Line undef)
// Falls eine fehlerhafte Zeile auftritt, wird eine Fehlermeldung nach cerr
// geschrieben und versucht aus den nicht fehlerhaften Bestandteilen ein
// hm_spec_line-Objekt zu generieren
// Falls das nicht gelingt, wird die Zeile ignoriert und mit der nchsten Zeile
// fortgefahren ohne vorher aus der Funktion zurckzukehren

bool hm_spec_scanner::Get (hm_spec_line& Line)
{
    // Anteil links von Doppelpunkt
    SkipWhitespaces (true);

    char c;
    if (!GetNext (c))
        return false;
    else
        Putback (c);

    string File;
    if (!ReadFile (File))
    {
        ReportError ("incomplete line");
        return d_Stream.good() ? Get (Line) : false;
    }

    hm_spec_line::file_type Type;
    if (!ReadType (Type))
    {
        ReportError ("incomplete line");
        return d_Stream.good() ? Get (Line) : false;
    }

    if (!GetNext (c) || c!=':')
    {
        SkipLine();
        ReportError ("incomplete line (expecting ':')");
        return d_Stream.good() ? Get (Line) : false;
    }

    Line = hm_spec_line (File, Type);

    // Anteil rechts von Doppelpunkt
    string Arg;
    while (!(Arg = ReadArg()).empty())
    {
        if (Type==hm_spec_line::NoFile &&
            find (d_LegalArgs.begin(), d_LegalArgs.end(), Arg)==d_LegalArgs.end())
            ReportError (Arg + " unknown");
        else
            Line.AddArg (Arg);
    }

    return true;
}
