//
// File:    ul_Scanner.c++
// Author:  Marc Zinnschlag

#ifndef HDR_ul_Scanner
#include "ul_Scanner.h"
#endif

// C

#ifndef LIB_assert
#include <cassert>
#define LIB_assert
#endif

// C++

#ifndef LIB_iostream
#include <iostream>
#define LIB_iostream
#endif

#ifndef HDR_ul_Token
#include "ul_Token.h"
#endif

// -----------------------------------------------------------------------------------------
// Klasse: ul_putback_buffer
// -----------------------------------------------------------------------------------------

class ul_putback_buffer
{
        ul_token *d_Array;
        int d_ArraySize;
        int d_Size;

        enum { InitSize = 4, GrowFactor = 2 };

    private:

    // Nicht implementiert

        ul_putback_buffer (const ul_putback_buffer&);
        ul_putback_buffer& operator= (const ul_putback_buffer&);

    public:

    // Konstruktoren

        ul_putback_buffer (int SizeHint);
        ~ul_putback_buffer();

    // Konstante

        int GetSize() const;
        // Anzahl Tokens in Putback-Buffer zurckgeben

        void Examine (ul_token *Token);
        // Token aus Putback-Buffer lesen ohne es zu Popen (undef, falls Buffer leer)

    // Nicht-Konstante

        void Push (const ul_token& Token);
        // Token in Putback-Buffer legen

        void Pop (ul_token *Token);
        // Token aus Putback-Buffer lesen (undef, falls Buffer leer)
};

// Konstruktoren ---------------------------------------------------------------------------

ul_putback_buffer::ul_putback_buffer (int SizeHint) : d_Size (0)
{
    d_ArraySize = SizeHint>0 ? SizeHint : InitSize;
    d_Array = new ul_token[d_ArraySize];
}

ul_putback_buffer::~ul_putback_buffer()
{
    delete [] d_Array;
}

// Konstante -------------------------------------------------------------------------------

// Anzahl Tokens in Putback-Buffer zurckgeben

int ul_putback_buffer::GetSize() const
{
    return d_Size;
}

// Token aus Putback-Buffer lesen ohne es zu Popen (undef, falls Buffer leer)

void ul_putback_buffer::Examine (ul_token *Token)
{
    assert (Token);
    assert (d_Size>0);
    *Token = d_Array[d_Size-1];
}

// Nicht-Konstante -------------------------------------------------------------------------

// Token in Putback-Buffer legen

void ul_putback_buffer::Push (const ul_token& Token)
{
    if (d_Size==d_ArraySize)
    {
        // Array vergrern
        ul_token *New = new ul_token[d_ArraySize *= GrowFactor];
        for (int i=0; i<d_Size; ++i)
            New[i] = d_Array[i];
        delete [] d_Array;
        d_Array = New;
    }

    d_Array[d_Size++] = Token;
}

// Token aus Putback-Buffer lesen (undef, falls Buffer leer)

void ul_putback_buffer::Pop (ul_token *Token)
{
    assert (Token);
    assert (d_Size>0);
    *Token = d_Array[--d_Size];
}

// -----------------------------------------------------------------------------------------
// Klasse: ul_scanner
// -----------------------------------------------------------------------------------------

// Private Hilfsfunktionen -----------------------------------------------------------------

bool ul_scanner::IsToBeIgnored (const ul_token& Token) const
{
    return !(1<<Token.GetType() & d_FilterFlags) && Token.GetType()!=ul_token::EndOfFile &&
        Token.GetType()!=ul_token::Empty;
}

void ul_scanner::MakeNewToken (ul_token *Token)
{
    assert (Token);

    if (GetPutbackSize())
    {
        d_PutbackBuffer->Pop (Token);
        return;
    }

    string String;

    int c;

    do
    {
        c = GetChar();
        int Result = InterpretChar (c, Token, String);
        if (Result>0)
        {
            if (Result>=1) d_State = WhitespaceState;
            if (Result>1 && c!=Endmark) d_Stream.putback ((char) c);
            break;
        }

        if (c!=Endmark && Result!=-1) String += c;
    }
    while (c!=Endmark);
}

// Falls Dateiende erreicht, wird Endmark zurckgegeben

int ul_scanner::GetChar()
{
    char c;
    if (!d_Stream.get (c)) return Endmark;
    return c;
}

// Return:
// -1 Token unvollstndig, Char nicht an *String anhngen
//  0 Token unvollstndig
//  1 Token vollstndig, Sprung zu WhitespaceState
//  2 Token vollstndig, letztes Zeichen zurcklegen, Sprung zu WhitespaceState

int ul_scanner::InterpretChar (int Char, ul_token *Token, string& String)
{
    assert (Token);

    switch (d_State)
    {
        case WhitespaceState:

            if (Char=='/')
                d_State = SlashState;
            else if (Char=='\'')
                d_State = CharState;
            else if (Char=='"')
            {
                d_State = StringState;
                return -1;
            }
            else if (Char>='1' && Char<='9')
                d_State = NumberState;
            else if (Char=='0')
                d_State = ZeroState;
            else if (Char=='.')
                d_State = PointState;
            else if (Char=='_' || Char>='a' && Char<='z' || Char>='A' && Char<='Z')
                d_State = NameState;
            else if (Char=='+' || Char=='-' || Char=='=' || Char=='!' || Char=='<' ||
                Char=='>' || Char==':' || Char=='&' || Char=='|')
                d_State = SpecialState;
            else if (Char==' ' || Char=='\n' || Char=='\t')
            {
                *Token = ul_token (ul_token::Whitespace, (char) Char);
                return 1;
            }
            else if (Char==Endmark)
            {
                *Token = ul_token (ul_token::EndOfFile);
                return 1;
            }
            else
            {
                *Token = ul_token (ul_token::Char, (char) Char);
                return 1;
            }

            break;

        case SlashState:

            if (Char=='/')
            {
                d_State = CppCommentState;
                String = "";
                return -1;
            }
            else if (Char=='*')
            {
                d_State = CCommentState;
                String = "";
                return -1;
            }
            else
            {
                *Token = ul_token (ul_token::Char, '/');
                return 2;
            }

            break;

        case CppCommentState:

            if (Char==Endmark || Char=='\n')
            {
                *Token = ul_token (ul_token::CppComment, String);
                return Char==Endmark ? 1 : 2;
            }

            break;

        case CCommentState:

            if (Char=='*')
            {
                d_State = CCommentEndState;
                return -1;
            }
            else if (Char==Endmark)
                d_State = CCommentEndState;

            break;

        case CCommentEndState:

            if (Char=='/' || Char==Endmark)
            {
                *Token = ul_token (ul_token::CComment, String);
                return 1;
            }
            else
                String += '*';

            break;

        case CharState:

            if (Char=='\\')
                d_State = CharBSState;
            else if (Char=='\n' || Char==Endmark)
            {
                *Token = ul_token (ul_token::GarbageCharLiteral, String);
                return 2;
            }
            else
                d_State = CharEndState;

            break;

        case CharBSState:

            if (Char=='\n' || Char==Endmark)
            {
                *Token = ul_token (ul_token::GarbageCharLiteral, String);
                return 2;
            }
            else
                d_State = CharEndState;

            break;

        case CharEndState:

            if (Char=='\'')
            {
                *Token = ul_token (ul_token::CharLiteral, String[String.length()-1]);
                return 1;
            }
            else
                d_State = CharGarbageState;

            break;

        case CharGarbageState:

            if (Char=='\'' || Char=='\n' || Char==Endmark)
            {
                if (Char=='\'') String += '\'';
                *Token = ul_token (ul_token::GarbageCharLiteral, String);
                return Char=='\'' ? 1 : 2;
            }

            break;

        case StringState:

            if (Char=='"')
            {
                *Token = ul_token (ul_token::String, String);
                return 1;
            }
            else if (Char=='\n' || Char==Endmark)
            {
                *Token = ul_token (ul_token::GarbageString, "\"" + String);
                return 2;
            }
            else if (Char=='\\')
                d_State = StringBSState;

            break;

        case StringBSState:

            if (Char==Endmark)
            {
                *Token = ul_token (ul_token::GarbageString, "\"" + String);
                return 2;
            }
            else
                d_State = StringState;

            break;

        case NumberState:

            if (Char=='e' || Char=='E')
                d_State = NumberEState;
            else if (Char=='.')
                d_State = PointState;
            else if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }

            break;

        case ZeroState:

            if (Char=='e' || Char=='E')
                d_State = NumberEState;
            else if (Char=='.')
                d_State = PointState;
            else if (Char=='x')
                d_State = XState;
            else if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }
            else
                d_State = NumberState;

            break;

        case PointState:

            if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }
            else
                d_State = Number2State;

            break;

        case NumberEState:

            if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::GarbageNumber, String);
                return 2;
            }
            else
                d_State = Number3State;

            break;

        case XState:

            if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }
            else
                d_State = NumberState;

            break;

        case Number2State:

            if (Char=='e' || Char=='E')
                d_State = NumberEState;
            else if (Char=='.')
            {
                *Token = ul_token (ul_token::GarbageNumber, String);
                return 2;
            }
            else if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }

            break;

        case Number3State:

            if (Char=='e' || Char=='E' ||Char=='.')
            {
                *Token = ul_token (ul_token::GarbageNumber, String);
                return 2;
            }
            else if (!(Char>='0' && Char<='9'))
            {
                *Token = ul_token (ul_token::Number, String);
                return 2;
            }

            break;

        case NameState:

            if (!(Char=='_' || Char>='0' && Char<='9' || Char>='a' && Char<='z' ||
                Char>='A' && Char<='Z'))
            {
                *Token = ul_token (ul_token::Name, String);
                return 2;
            }

            break;

        case SpecialState:
        {
            int i=0;
            for (; i<ul_token::SpecialIndices; ++i)
                if (String[0]==ul_token::GetSpecialString (i)[0] &&
                    (char) Char==ul_token::GetSpecialString (i)[1])
                    break;

            if (i<ul_token::SpecialIndices)
            {
                *Token = ul_token (ul_token::Special, (char) i);
                return 1;
            }
            else
            {
                *Token = ul_token (ul_token::Char, String[0]);
                return 2;
            }

            break;
        }
    }

    return 0;
}

// Konstruktoren ---------------------------------------------------------------------------

ul_scanner::ul_scanner (istream& Stream, int PutbackSizeHint, bool IgnorePreprocessor)
: d_Stream (Stream), d_PutbackBuffer (new ul_putback_buffer (PutbackSizeHint)),
  d_EOF (false), d_State (WhitespaceState), d_FilterFlags (~0),
  d_IgnorePreprocessor (IgnorePreprocessor)
{}

ul_scanner::~ul_scanner()
{
    delete d_PutbackBuffer;
}

// Konstante -------------------------------------------------------------------------------

// Anzahl Tokens in Putback-Buffer zurckgeben

int ul_scanner::GetPutbackSize() const
{
    return d_PutbackBuffer->GetSize();
}

// Ende des Files noch nicht erreicht?

ul_scanner::operator const void *() const
{
    return IsEOF() ? 0 : this;
}

// Nicht-Konstante -------------------------------------------------------------------------

// Nchstes Token lesen
// Falls ein EOF-Token zurckgegeben wird, wird fr alle weiteren Aufrufe
// dieser Funktion dasselbe Token zurckgegeben
// Return: 0 EOF-Token erreicht, 1 anderes Token

int ul_scanner::GetNext (ul_token *Token)
{
    assert (Token);

    if (d_EOF)
    {
        *Token = ul_token (ul_token::EndOfFile);
        return 0;
    }

    do
    {
        MakeNewToken (Token);

        if (d_IgnorePreprocessor && *Token==ul_token (ul_token::Char, '#'))
        {
            for (ul_token Prev; *this; MakeNewToken (Token))
            {
                if (*Token==ul_token (ul_token::Whitespace, '\n') &&
                    Prev!=ul_token (ul_token::Char, '\\'))
                    break;

                Prev = *Token;
            }
        }
    }
    while (IsToBeIgnored (*Token));

    if (*Token==ul_token (ul_token::EndOfFile))
        d_EOF = 1;

    return !d_EOF;
}

// Token zurckliegen (mu nicht unbedingt dem aus GetNext enthaltenen Token entsprechend)

void ul_scanner::Putback (const ul_token& Token)
{
    if (!IsToBeIgnored (Token))
        d_PutbackBuffer->Push (Token);
}

// Angegebenen TokenType zulassen
// Alle Types bei Konstruktion zugelassen

void ul_scanner::Enable (ul_token::type TokenType)
{
    assert (TokenType>=0 && TokenType<31);
    d_FilterFlags |= 1<<TokenType;
}

// Angegebenen TokenType unterdrcken
// Tokens der Typen EndOfFile und Empty knnen nicht unterdrckt werden (Aufruf von Disable
// fr diese Typen wird ignoriert)

void ul_scanner::Disable (ul_token::type TokenType)
{
    assert (TokenType>=0 && TokenType<31);
    d_FilterFlags &= ~(1<<TokenType);
}

// Freie Operatoren ------------------------------------------------------------------------

ul_scanner& operator<< (ul_scanner& Scanner, const ul_token& Token)
{
    Scanner.Putback (Token);
    return Scanner;
}

ul_scanner& operator>> (ul_scanner& Scanner, ul_token& Token)
{
    Scanner.GetNext (&Token);
    return Scanner;
}
